@@ -1,5 +1,3 @@
-#!/usr/bin/python
-
 from __future__ import unicode_literals
 
 import base64
@@ -8,7 +6,7 @@ import os
 import socket
 import sys
 import threading
-from wsgiref.simple_server import make_server, WSGIRequestHandler
+from wsgiref.simple_server import make_server, WSGIServer, WSGIRequestHandler
 
 from .openmetrics import exposition as openmetrics
 from .registry import REGISTRY
@@ -27,27 +25,35 @@ except ImportError:
     from urllib.request import build_opener, Request, HTTPHandler
     from urllib.parse import quote_plus, parse_qs, urlparse
 
-
 CONTENT_TYPE_LATEST = str('text/plain; version=0.0.4; charset=utf-8')
-'''Content type of the latest text format'''
+"""Content type of the latest text format"""
 
 PYTHON26_OR_OLDER = sys.version_info < (2, 7)
+PYTHON376_OR_NEWER = sys.version_info > (3, 7, 5)
+
+
+def _bake_output(registry, accept_header, params):
+    """Bake output for metrics output."""
+    encoder, content_type = choose_encoder(accept_header)
+    if 'name[]' in params:
+        registry = registry.restricted_registry(params['name[]'])
+    output = encoder(registry)
+    return str('200 OK'), (str('Content-Type'), content_type), output
 
 
 def make_wsgi_app(registry=REGISTRY):
-    '''Create a WSGI app which serves the metrics from a registry.'''
-    def prometheus_app(environ, start_response):
-        params = parse_qs(environ.get('QUERY_STRING', ''))
-        r = registry
-        encoder, content_type = choose_encoder(environ.get('HTTP_ACCEPT'))
-        if 'name[]' in params:
-            r = r.restricted_registry(params['name[]'])
-        output = encoder(r)
+    """Create a WSGI app which serves the metrics from a registry."""
 
-        status = str('200 OK')
-        headers = [(str('Content-type'), content_type)]
-        start_response(status, headers)
+    def prometheus_app(environ, start_response):
+        # Prepare parameters
+        accept_header = environ.get('HTTP_ACCEPT')
+        params = parse_qs(environ.get('QUERY_STRING', ''))
+        # Bake output
+        status, header, output = _bake_output(registry, accept_header, params)
+        # Return output
+        start_response(status, [header])
         return [output]
+
     return prometheus_app
 
 
@@ -58,66 +64,84 @@ class _SilentHandler(WSGIRequestHandler):
         """Log nothing."""
 
 
+class ThreadingWSGIServer(ThreadingMixIn, WSGIServer):
+    """Thread per request HTTP server."""
+    # Make worker threads "fire and forget". Beginning with Python 3.7 this
+    # prevents a memory leak because ``ThreadingMixIn`` starts to gather all
+    # non-daemon threads in a list in order to join on them at server close.
+    daemon_threads = True
+
+
 def start_wsgi_server(port, addr='', registry=REGISTRY):
     """Starts a WSGI server for prometheus metrics as a daemon thread."""
     app = make_wsgi_app(registry)
-    httpd = make_server(addr, port, app, handler_class=_SilentHandler)
+    httpd = make_server(addr, port, app, ThreadingWSGIServer, handler_class=_SilentHandler)
     t = threading.Thread(target=httpd.serve_forever)
     t.daemon = True
     t.start()
 
 
-def generate_latest(registry=REGISTRY):
-    '''Returns the metrics from the registry in latest text format as a string.'''
+start_http_server = start_wsgi_server
 
-    def sample_line(s):
-        if s.labels:
+
+def generate_latest(registry=REGISTRY):
+    """Returns the metrics from the registry in latest text format as a string."""
+
+    def sample_line(line):
+        if line.labels:
             labelstr = '{{{0}}}'.format(','.join(
                 ['{0}="{1}"'.format(
-                 k, v.replace('\\', r'\\').replace('\n', r'\n').replace('"', r'\"'))
-                 for k, v in sorted(s.labels.items())]))
+                    k, v.replace('\\', r'\\').replace('\n', r'\n').replace('"', r'\"'))
+                    for k, v in sorted(line.labels.items())]))
         else:
             labelstr = ''
         timestamp = ''
-        if s.timestamp is not None:
+        if line.timestamp is not None:
             # Convert to milliseconds.
-            timestamp = ' {0:d}'.format(int(float(s.timestamp) * 1000))
+            timestamp = ' {0:d}'.format(int(float(line.timestamp) * 1000))
         return '{0}{1} {2}{3}\n'.format(
-            s.name, labelstr, floatToGoString(s.value), timestamp)
+            line.name, labelstr, floatToGoString(line.value), timestamp)
 
     output = []
     for metric in registry.collect():
-        mname = metric.name
-        mtype = metric.type
-        # Munging from OpenMetrics into Prometheus format.
-        if mtype == 'counter':
-            mname = mname + '_total'
-        elif mtype == 'info':
-            mname = mname + '_info'
-            mtype = 'gauge'
-        elif mtype == 'stateset':
-            mtype = 'gauge'
-        elif mtype == 'gaugehistogram':
-            # A gauge histogram is really a gauge,
-            # but this captures the strucutre better.
-            mtype = 'histogram'
-        elif mtype == 'unknown':
-            mtype = 'untyped'
+        try:
+            mname = metric.name
+            mtype = metric.type
+            # Munging from OpenMetrics into Prometheus format.
+            if mtype == 'counter':
+                mname = mname + '_total'
+            elif mtype == 'info':
+                mname = mname + '_info'
+                mtype = 'gauge'
+            elif mtype == 'stateset':
+                mtype = 'gauge'
+            elif mtype == 'gaugehistogram':
+                # A gauge histogram is really a gauge,
+                # but this captures the structure better.
+                mtype = 'histogram'
+            elif mtype == 'unknown':
+                mtype = 'untyped'
 
-        output.append('# HELP {0} {1}\n'.format(
-            mname, metric.documentation.replace('\\', r'\\').replace('\n', r'\n')))
-        output.append('# TYPE {0} {1}\n'.format(mname, mtype))
+            output.append('# HELP {0} {1}\n'.format(
+                mname, metric.documentation.replace('\\', r'\\').replace('\n', r'\n')))
+            output.append('# TYPE {0} {1}\n'.format(mname, mtype))
+
+            om_samples = {}
+            for s in metric.samples:
+                for suffix in ['_created', '_gsum', '_gcount']:
+                    if s.name == metric.name + suffix:
+                        # OpenMetrics specific sample, put in a gauge at the end.
+                        om_samples.setdefault(suffix, []).append(sample_line(s))
+                        break
+                else:
+                    output.append(sample_line(s))
+        except Exception as exception:
+            exception.args = (exception.args or ('',)) + (metric,)
+            raise
 
-        om_samples = {}
-        for s in metric.samples:
-            for suffix in ['_created', '_gsum', '_gcount']:
-                if s.name == metric.name + suffix:
-                    # OpenMetrics specific sample, put in a gauge at the end.
-                    om_samples.setdefault(suffix, []).append(sample_line(s))
-                    break
-            else:
-                output.append(sample_line(s))
         for suffix, lines in sorted(om_samples.items()):
+            output.append('# HELP {0}{1} {2}\n'.format(metric.name, suffix, 
+                                                       metric.documentation.replace('\\', r'\\').replace('\n', r'\n')))
             output.append('# TYPE {0}{1} gauge\n'.format(metric.name, suffix))
             output.extend(lines)
     return ''.join(output).encode('utf-8')
@@ -129,7 +153,7 @@ def choose_encoder(accept_header):
         if accepted.split(';')[0].strip() == 'application/openmetrics-text':
             return (openmetrics.generate_latest,
                     openmetrics.CONTENT_TYPE_LATEST)
-    return (generate_latest, CONTENT_TYPE_LATEST)
+    return generate_latest, CONTENT_TYPE_LATEST
 
 
 class MetricsHandler(BaseHTTPRequestHandler):
@@ -137,18 +161,15 @@ class MetricsHandler(BaseHTTPRequestHandler):
     registry = REGISTRY
 
     def do_GET(self):
+        # Prepare parameters
         registry = self.registry
+        accept_header = self.headers.get('Accept')
         params = parse_qs(urlparse(self.path).query)
-        encoder, content_type = choose_encoder(self.headers.get('Accept'))
-        if 'name[]' in params:
-            registry = registry.restricted_registry(params['name[]'])
-        try:
-            output = encoder(registry)
-        except:
-            self.send_error(500, 'error generating metric output')
-            raise
-        self.send_response(200)
-        self.send_header('Content-Type', content_type)
+        # Bake output
+        status, header, output = _bake_output(registry, accept_header, params)
+        # Return output
+        self.send_response(int(status.split(' ')[0]))
+        self.send_header(*header)
         self.end_headers()
         self.wfile.write(output)
 
@@ -171,24 +192,11 @@ class MetricsHandler(BaseHTTPRequestHandler):
         return MyMetricsHandler
 
 
-class _ThreadingSimpleServer(ThreadingMixIn, HTTPServer):
-    """Thread per request HTTP server."""
-
-
-def start_http_server(port, addr='', registry=REGISTRY):
-    """Starts an HTTP server for prometheus metrics as a daemon thread"""
-    CustomMetricsHandler = MetricsHandler.factory(registry)
-    httpd = _ThreadingSimpleServer((addr, port), CustomMetricsHandler)
-    t = threading.Thread(target=httpd.serve_forever)
-    t.daemon = True
-    t.start()
-
-
 def write_to_textfile(path, registry):
-    '''Write metrics to the given path.
+    """Write metrics to the given path.
 
     This is intended for use with the Node exporter textfile collector.
-    The path must end in .prom for the textfile collector to process it.'''
+    The path must end in .prom for the textfile collector to process it."""
     tmppath = '%s.%s.%s' % (path, os.getpid(), threading.current_thread().ident)
     with open(tmppath, 'wb') as f:
         f.write(generate_latest(registry))
@@ -197,9 +205,10 @@ def write_to_textfile(path, registry):
 
 
 def default_handler(url, method, timeout, headers, data):
-    '''Default handler that implements HTTP/HTTPS connections.
+    """Default handler that implements HTTP/HTTPS connections.
+
+    Used by the push_to_gateway functions. Can be re-used by other handlers."""
 
-    Used by the push_to_gateway functions. Can be re-used by other handlers.'''
     def handle():
         request = Request(url, data=data)
         request.get_method = lambda: method
@@ -214,13 +223,14 @@ def default_handler(url, method, timeout, headers, data):
 
 
 def basic_auth_handler(url, method, timeout, headers, data, username=None, password=None):
-    '''Handler that implements HTTP/HTTPS connections with Basic Auth.
+    """Handler that implements HTTP/HTTPS connections with Basic Auth.
 
     Sets auth headers using supplied 'username' and 'password', if set.
-    Used by the push_to_gateway functions. Can be re-used by other handlers.'''
+    Used by the push_to_gateway functions. Can be re-used by other handlers."""
+
     def handle():
-        '''Handler that implements HTTP Basic Auth.
-        '''
+        """Handler that implements HTTP Basic Auth.
+        """
         if username is not None and password is not None:
             auth_value = '{0}:{1}'.format(username, password).encode('utf-8')
             auth_token = base64.b64encode(auth_value)
@@ -234,7 +244,7 @@ def basic_auth_handler(url, method, timeout, headers, data, username=None, passw
 def push_to_gateway(
         gateway, job, registry, grouping_key=None, timeout=30,
         handler=default_handler):
-    '''Push metrics to the given pushgateway.
+    """Push metrics to the given pushgateway.
 
     `gateway` the url for your push gateway. Either of the form
               'http://pushgateway.local', or 'pushgateway.local'.
@@ -271,14 +281,14 @@ def push_to_gateway(
               Message Body.
 
     This overwrites all metrics with the same job and grouping_key.
-    This uses the PUT HTTP method.'''
+    This uses the PUT HTTP method."""
     _use_gateway('PUT', gateway, job, registry, grouping_key, timeout, handler)
 
 
 def pushadd_to_gateway(
         gateway, job, registry, grouping_key=None, timeout=30,
         handler=default_handler):
-    '''PushAdd metrics to the given pushgateway.
+    """PushAdd metrics to the given pushgateway.
 
     `gateway` the url for your push gateway. Either of the form
               'http://pushgateway.local', or 'pushgateway.local'.
@@ -297,13 +307,13 @@ def pushadd_to_gateway(
               for implementation requirements.
 
     This replaces metrics with the same name, job and grouping_key.
-    This uses the POST HTTP method.'''
+    This uses the POST HTTP method."""
     _use_gateway('POST', gateway, job, registry, grouping_key, timeout, handler)
 
 
 def delete_from_gateway(
         gateway, job, grouping_key=None, timeout=30, handler=default_handler):
-    '''Delete metrics from the given pushgateway.
+    """Delete metrics from the given pushgateway.
 
     `gateway` the url for your push gateway. Either of the form
               'http://pushgateway.local', or 'pushgateway.local'.
@@ -321,15 +331,19 @@ def delete_from_gateway(
               for implementation requirements.
 
     This deletes metrics with the given job and grouping_key.
-    This uses the DELETE HTTP method.'''
+    This uses the DELETE HTTP method."""
     _use_gateway('DELETE', gateway, job, None, grouping_key, timeout, handler)
 
 
 def _use_gateway(method, gateway, job, registry, grouping_key, timeout, handler):
     gateway_url = urlparse(gateway)
-    if not gateway_url.scheme or (PYTHON26_OR_OLDER and gateway_url.scheme not in ['http', 'https']):
+    # See https://bugs.python.org/issue27657 for details on urlparse in py>=3.7.6.
+    if not gateway_url.scheme or (
+            (PYTHON376_OR_NEWER or PYTHON26_OR_OLDER)
+            and gateway_url.scheme not in ['http', 'https']
+    ):
         gateway = 'http://{0}'.format(gateway)
-    url = '{0}/metrics/job/{1}'.format(gateway, quote_plus(job))
+    url = '{0}/metrics/{1}/{2}'.format(gateway, *_escape_grouping_key("job", job))
 
     data = b''
     if method != 'DELETE':
@@ -338,7 +352,7 @@ def _use_gateway(method, gateway, job, registry, grouping_key, timeout, handler)
     if grouping_key is None:
         grouping_key = {}
     url += ''.join(
-        '/{0}/{1}'.format(quote_plus(str(k)), quote_plus(str(v)))
+        '/{0}/{1}'.format(*_escape_grouping_key(str(k), str(v)))
         for k, v in sorted(grouping_key.items()))
 
     handler(
@@ -347,8 +361,26 @@ def _use_gateway(method, gateway, job, registry, grouping_key, timeout, handler)
     )()
 
 
+def _escape_grouping_key(k, v):
+    if v == "" :
+        # Per https://github.com/prometheus/pushgateway/pull/346.
+        return k + "@base64", "="
+    elif '/' in v:
+        # Added in Pushgateway 0.9.0.
+        return k + "@base64", base64.urlsafe_b64encode(v.encode("utf-8")).decode("utf-8")
+    else:
+        return k, quote_plus(v)
+
+
 def instance_ip_grouping_key():
-    '''Grouping key with instance set to the IP Address of this host.'''
+    """Grouping key with instance set to the IP Address of this host."""
     with closing(socket.socket(socket.AF_INET, socket.SOCK_DGRAM)) as s:
         s.connect(('localhost', 0))
         return {'instance': s.getsockname()[0]}
+
+
+try:
+    # Python >3.5 only
+    from .asgi import make_asgi_app
+except:
+    pass
